-- scope stuff
qualityoflife = qualityoflife or {};
local q = qualityoflife;

q.modules = q.modules or {};
q.modules["QoL_ClassicThreat-TBC"] = q.modules["QoL_ClassicThreat-TBC"] or {};
local c = q.modules["QoL_ClassicThreat-TBC"];

local bars, barFrame = {};

local minWidth, minHeight = 100, 50;
local barHeight, barMarginX, barMarginY = 12, 4, 2;
local npcColor = { r = 1, g = 0, b = 0, a = 1 };

local function GetBarByGuid(guid)
    for _, bar in ipairs(bars) do
        if bar.guid == guid then
            return bar;
        end
    end
end

local function GetBar(guid)
    local barFrame, bar = c:GetBarFrame(), GetBarByGuid(guid);
    if not bar then

        -- try to get existing, inactive bar (performance optimization)
        for _, b in ipairs(bars) do
            if not b:IsActive() then
                bar = b;
                break;
            end
        end

        if not bar then
            -- create new bar if needed
            local index = table.getn(bars);
            bar = CreateFrame("Frame", "ClassicThreat_BarFrame_Bar" .. index, barFrame, BackdropTemplateMixin and "BackdropTemplate");
            bar:SetPoint("TOPLEFT", barFrame, "TOPLEFT", 0, - 2 - barFrame.header:GetHeight());
            bar:SetBackdrop({
                bgFile = [[Interface\Buttons\WHITE8x8]]
            });

            bar.text = bar:CreateFontString("ClassicThreat_BarFrame_Bar" .. index .. "_FontString", "ARTWORK", "GameFontNormalSmall");
            bar.text:SetTextColor(1, 1, 1, 1);
            bar.text:SetPoint("TOPLEFT", bar, 4, 0);
            bar.text:SetPoint("BOTTOMLEFT", bar, 4, 0);

            bar.SetThreatInfo = function (self, threatInfo)
                self.threatInfo = threatInfo;
            end

            bar.SetGuid = function (self, guid)
                self.guid = guid;
            end

            bar.SetColor = function (self, colorTable)
                local colors = {};
                if colorTable.r and colors.g and colors.b then
                    -- named properties
                    colors = colorTable;
                else
                    -- simple array
                    colors.r = colorTable[1];
                    colors.g = colorTable[2];
                    colors.b = colorTable[3];
                    colors.a = colorTable[4];
                end
                colors.a = colors.a or 1;
                self:SetBackdropColor(colors.r, colors.g, colors.b, colors.a);
            end

            bar.SetText = function (self, text)
                self.text:SetText(text);
            end

            bar.SetActive = function(self, value)
                self.isActive = value;
            end

            bar.IsActive = function(self)
                return self.isActive;
            end

            bar.SetPending = function (self, value)
                self.isPending = value;
            end

            bar.IsPending = function (self)
                return self.isPending;
            end

            bar.Move = function (self, targetTop, targetWidth)
                -- TODO (optional): smooth movement
                self:ClearAllPoints();
                self:SetPoint("TOPLEFT", c:GetBarFrame(), "TOPLEFT", barMarginX, -targetTop);
                self:SetSize(targetWidth, barHeight);
            end

            bar.Reset = function(self)
                self:SetPending(false);
                self:SetActive(false);
                self:SetText("");
                self:SetColor({ r = 1, g = 1, b = 1, a = 1 });
                self:SetThreatInfo(nil);
                self:SetGuid(nil);
                self:Hide();
            end

            tinsert(bars, bar);
        end
    end

    bar:SetGuid(guid);
    return bar;
end

local function CalculateWidth(threatPct)
    local maxWidth = c:GetBarFrame():GetWidth() - 2 * barMarginX;
    local width = threatPct * maxWidth / 100;
    if width > maxWidth then
        return maxWidth;
    end
    return width;
end

local function UpdateBars(barFrame, threatCache)
    -- set all bars as inactive (some will be set to active in this process)
    for _, bar in ipairs(bars) do
        bar:SetPending(true);
    end

    if threatCache then
        local activeBars = {};

        if QOL_OPTIONS[c.name][c.BARFRAME]["mode"] == c.BARFRAME_MODE_UNIT then
            -- unit mode

            local unitGuid = UnitGUID(QOL_OPTIONS[c.name][c.BARFRAME]["unit"]);
            if unitGuid and threatCache[unitGuid] then
                for targetGuid, threatInfo in pairs(threatCache[unitGuid]) do
                    -- get bar and set data
                    local bar = GetBar(targetGuid);
                    bar:SetActive(true);
                    bar:SetPending(false);
                    bar:SetColor(c:GetThreatStatusColors()[threatInfo.status]); -- default npc color
                    bar:SetText(q:MathRound(threatInfo.rawthreatpct, 0) .. "% - " ..threatInfo.targetName);
                    bar:SetThreatInfo(threatInfo);
                    bar:Show();

                    -- prepare bar sorting
                    activeBars[threatInfo.threat] = activeBars[threatInfo.threat] or {};
                    tinsert(activeBars[threatInfo.threat], bar);
                end
            end

        elseif QOL_OPTIONS[c.name][c.BARFRAME]["mode"] == c.BARFRAME_MODE_RAID then
            -- raid mode

            local unitGuid = UnitGUID("target");
            --loop players for this target
            -- display
            --RAID_CLASS_COLORS[select(2, UnitClass(unitId))];

        end

        -- sort and align bars
        local headerHeight = barFrame.header:GetHeight();
        local i, sortedBarKeys = 0, q:GetSortedTableKeys(activeBars, true);
        for _, key in ipairs(sortedBarKeys) do
            for _, bar in ipairs(activeBars[key]) do
                local top = headerHeight + i * (barHeight + barMarginY);
                bar:Move(top, CalculateWidth(bar.threatInfo.threat));
                i = i + 1;
            end
        end
    end

    -- reset/hide inactive bars
    for _, bar in ipairs(bars) do
        if bar:IsPending() then
            bar:Reset();
        end
    end
end

function c:GetBarFrame()
    if not barFrame then
        -- frame
        barFrame = CreateFrame("Frame", "ClassicThreat_BarFrame", UIParent, BackdropTemplateMixin and "BackdropTemplate");
        barFrame:SetFrameStrata("HIGH");
        barFrame:SetMinResize(minWidth, minHeight);

        barFrame:SetBackdrop({
            edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
            tile = true,
            tileSize = 10,
            edgeSize = 10,
            insets = {left = 4, right = 4, top = 4, bottom = 4}
        });

        barFrame:SetClipsChildren(true);
        barFrame:SetResizable(true);
        barFrame:SetMovable(true);
        barFrame:EnableMouse(true);
        barFrame:RegisterForDrag("LeftButton");
        barFrame:SetScript("OnDragStart", function (self)
            self:StartMoving();
        end);
        barFrame:SetScript("OnDragStop", function (self)
            self:StopMovingOrSizing();
            QOL_OPTIONS[c.name][c.BARFRAME]["bottom"] = self:GetBottom();
            QOL_OPTIONS[c.name][c.BARFRAME]["left"] = self:GetLeft();
        end);
        barFrame:SetUserPlaced(true);

        -- header
        barFrame.header = CreateFrame("Frame", "ClassicThreat_BarFrame_Header", barFrame, BackdropTemplateMixin and "BackdropTemplate");
        barFrame.header:SetPoint("TOPLEFT", barFrame, 6, 0);
        barFrame.header:SetPoint("TOPRIGHT", barFrame, 6, 0);
        barFrame.header:SetHeight(20);

        barFrame.header.text = barFrame.header:CreateFontString( "ClassicThreat_BarFrame_Header_FontString", "ARTWORK", "GameFontNormalSmall");
        barFrame.header.text:SetTextColor(1, 1, 1, 1);
        barFrame.header.text:SetPoint("TOPLEFT", barFrame.header);
        barFrame.header.text:SetPoint("BOTTOMLEFT", barFrame.header);
        barFrame.header.text:SetText("ClassicThreat");

        -- data
        barFrame.data = CreateFrame("Frame", "ClassicThreat_BarFrame_Data", barFrame, BackdropTemplateMixin and "BackdropTemplate");
        barFrame.data:SetPoint("TOPLEFT", barFrame, 0, 20);
        barFrame.data:SetPoint("BOTTOMRIGHT", barFrame);

        -- close button
        barFrame.closeButton = CreateFrame("Button", "ClassicThreat_BarFrame_CloseButton", barFrame, "UIPanelCloseButton");
        barFrame.closeButton:SetSize(20, 20);
        barFrame.closeButton:SetPoint("TOPRIGHT");
        barFrame.closeButton:SetScript("OnClick", function ()
            c:ToggleBarFrame(false);
        end);

        -- resize button
        barFrame.resizeButton = CreateFrame("Button", "ClassicThreat_BarFrame_ResizeButton", barFrame);
        barFrame.resizeButton:SetSize(19, 15);
        barFrame.resizeButton:SetPoint("BOTTOMRIGHT");
        barFrame.resizeButton:SetScript("OnMouseDown", function (self)
            barFrame:StartSizing();
        end);
        barFrame.resizeButton:SetScript("OnMouseUp", function(self)
            barFrame:StopMovingOrSizing();
            QOL_OPTIONS[c.name][c.BARFRAME]["width"] = barFrame:GetWidth();
            QOL_OPTIONS[c.name][c.BARFRAME]["height"] = barFrame:GetHeight();
            QOL_OPTIONS[c.name][c.BARFRAME]["bottom"] = barFrame:GetBottom();
        end)

        barFrame.Update = function (self, threatCache)
            UpdateBars(self, threatCache);
        end

        -- additional
        q:HandleElvUi(barFrame, function() end);
    end
    return barFrame;
end

function c:ToggleBarFrame(value)
    local barFrame = c:GetBarFrame();
    if value == nil then
       value = not barFrame:IsVisible();
    end;

    QOL_OPTIONS[c.name][c.BARFRAME][c.OPT_ENABLED] = value;
    if value then
        barFrame:ClearAllPoints();
        barFrame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", QOL_OPTIONS[c.name][c.BARFRAME]["left"], QOL_OPTIONS[c.name][c.BARFRAME]["bottom"]);
        local width, height = QOL_OPTIONS[c.name][c.BARFRAME]["width"], QOL_OPTIONS[c.name][c.BARFRAME]["height"];
        if width < minWidth then
            width = minWidth;
        end
        if height < minHeight then
            height = minHeight;
        end
        barFrame:SetSize(width, height);
        barFrame:Show();
    else
        barFrame:Hide();
    end
end

function c:UpdateBarFrame(threatCache)
    local barFrame = c:GetBarFrame();
    if barFrame:IsVisible() then
        barFrame:Update(threatCache);
    end
end

function c:ResetBarFrame()
    c:GetBarFrame():Update();
end

local function ChangeTestBars(mockThreatCache, testedUnitGuid)
    for k, v in pairs(mockThreatCache[testedUnitGuid]) do
        v.threat = random(0, 100);
    end
    UpdateBars(c:GetBarFrame(), mockThreatCache);

    C_Timer.After(3, function ()
        ChangeTestBars(mockThreatCache, testedUnitGuid);
    end);
end

function c:TestBars()
    c:ToggleBarFrame(true);
    local mockThreatCache, testedUnitGuid = {}, UnitGUID(QOL_OPTIONS[c.name][c.BARFRAME]["unit"]);
    mockThreatCache[testedUnitGuid] = mockThreatCache[testedUnitGuid] or {};

    for i = 0, 2 do
        local threatInfo = {};
        threatInfo.testedUnitId = "testedUnitId";
        threatInfo.targetUnitId = i;
        threatInfo.targetGuid = "targetGuid";
        threatInfo.targetName = i;
        threatInfo.isTanking = "isTanking";
        threatInfo.status = "status";
        mockThreatCache[testedUnitGuid]["targetGuid" .. i] = threatInfo;
    end
    ChangeTestBars(mockThreatCache, testedUnitGuid);
end